import { EMPTY_OBJECT_TYPE, UNKNOWN_TYPE, dereferenceType, objectType, maybeOptionalType, helpers, NODE_TYPE, functionType, arrayType, createTypeAnalyzer } from '@previewjs/type-analyzer';
import path from 'path';
import ts from 'typescript';
import url from 'url';
import { generatePreviewableId, decodePreviewableId } from '@previewjs/analyzer-api';
import { parseSerializableValue } from '@previewjs/serializable-values';
import { extractStoriesInfo, extractArgs, extractCsf3Stories, resolvePreviewableId } from '@previewjs/storybook-helpers';

function analyze(logger, typeResolver, signature) {
  const firstParam = signature.getParameters()[0];
  if (!firstParam) {
    return {
      props: EMPTY_OBJECT_TYPE,
      types: {}
    };
  }
  if (!firstParam.valueDeclaration) {
    return {
      props: UNKNOWN_TYPE,
      types: {}
    };
  }
  const type = typeResolver.checker.getTypeOfSymbolAtLocation(
    firstParam,
    firstParam.valueDeclaration
  );
  try {
    let { type: props, collected: types } = typeResolver.resolveType(type);
    [props] = dereferenceType(props, types, []);
    stripUnusedProps:
      if (props.kind === "object" && ts.isParameter(firstParam.valueDeclaration)) {
        if (ts.isObjectBindingPattern(firstParam.valueDeclaration.name)) {
          const bindingPattern = firstParam.valueDeclaration.name;
          const usedProps = /* @__PURE__ */ new Set();
          const propsWithDefault = /* @__PURE__ */ new Set();
          for (const element of bindingPattern.elements) {
            if (element.dotDotDotToken) {
              break stripUnusedProps;
            }
            const elementName = element.propertyName || element.name;
            if (!ts.isIdentifier(elementName)) {
              break stripUnusedProps;
            }
            usedProps.add(elementName.text);
            if (element.initializer) {
              propsWithDefault.add(elementName.text);
            }
          }
          props = objectType(
            Object.fromEntries(
              Object.entries(props.fields).filter(([key]) => usedProps.has(key)).map(([key, type2]) => [
                key,
                maybeOptionalType(type2, propsWithDefault.has(key))
              ])
            )
          );
        }
      }
    if (props.kind === "object") {
      props = {
        kind: "object",
        fields: Object.fromEntries(
          Object.entries(props.fields).filter(
            ([key]) => !["jsx", "key", "ref"].includes(key)
          )
        )
      };
    }
    return { props, types };
  } catch (e) {
    logger.warn(
      `Unable to resolve props type for ${typeResolver.checker.typeToString(
        type
      )}`,
      e
    );
  }
  return {
    props: UNKNOWN_TYPE,
    types: {}
  };
}

async function crawlFile(logger, resolver, rootDir, absoluteFilePath) {
  const sourceFile = resolver.sourceFile(absoluteFilePath);
  if (!sourceFile) {
    return [];
  }
  const functions = [];
  for (const statement of sourceFile.statements) {
    if (ts.isExportAssignment(statement)) {
      if (ts.isIdentifier(statement.expression)) {
        continue;
      }
      functions.push(["default", statement, statement.expression]);
    } else if (ts.isVariableStatement(statement)) {
      for (const declaration of statement.declarationList.declarations) {
        if (!ts.isIdentifier(declaration.name) || !declaration.initializer) {
          continue;
        }
        functions.push([
          declaration.name.text,
          statement,
          declaration.initializer
        ]);
      }
    } else if (ts.isFunctionDeclaration(statement)) {
      const isDefaultExport = !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.ExportKeyword
      ) && !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.DefaultKeyword
      );
      const name = statement.name?.text;
      if (isDefaultExport || name) {
        functions.push([name || "default", statement, statement]);
      }
    } else if (ts.isClassDeclaration(statement) && statement.name) {
      functions.push([statement.name.text, statement, statement]);
    }
  }
  const storiesInfo = extractStoriesInfo(sourceFile);
  const previewables = [];
  const args = extractArgs(sourceFile);
  const nameToExportedName = helpers.extractExportedNames(sourceFile);
  async function extractPreviewable(basePreviewable, node, name) {
    if (name === "default" && storiesInfo) {
      return null;
    }
    const storyArgs = args[name];
    const isExported = name === "default" || !!nameToExportedName[name];
    const signature = extractComponentSignature(resolver.checker, node);
    if (storiesInfo && isExported && (storyArgs || signature?.parameters.length === 0)) {
      const associatedComponent = extractStoryAssociatedComponent(
        logger,
        resolver,
        rootDir,
        storiesInfo.component
      );
      return {
        ...basePreviewable,
        analyze: async () => ({
          args: storyArgs ? {
            sourcePosition: {
              start: storyArgs.getStart(),
              end: storyArgs.getEnd()
            },
            value: await parseSerializableValue(storyArgs)
          } : null
        }),
        associatedComponent
      };
    }
    if (signature) {
      return {
        ...basePreviewable,
        exported: isExported,
        analyze: async () => analyze(logger, resolver, signature)
      };
    }
    return null;
  }
  for (const [name, statement, node] of functions) {
    const previewable = await extractPreviewable(
      {
        id: generatePreviewableId({
          filePath: path.relative(rootDir, absoluteFilePath),
          name
        }),
        sourcePosition: {
          start: statement.getStart(),
          end: statement.getEnd()
        }
      },
      node,
      name
    );
    if (previewable) {
      previewables.push(previewable);
    }
  }
  return [
    ...previewables,
    ...await extractCsf3Stories(rootDir, resolver, sourceFile, async (id) => {
      const { filePath } = decodePreviewableId(id);
      const component = (await crawlFile(logger, resolver, rootDir, path.join(rootDir, filePath))).find((c) => c.id === id);
      if (!component || !("exported" in component)) {
        return {
          props: UNKNOWN_TYPE,
          types: {}
        };
      }
      return component.analyze();
    })
  ];
}
function extractStoryAssociatedComponent(logger, resolver, rootDir, component) {
  const resolvedStoriesPreviewableId = resolvePreviewableId(
    rootDir,
    resolver.checker,
    component
  );
  return component && resolvedStoriesPreviewableId ? {
    id: resolvedStoriesPreviewableId,
    analyze: async () => {
      const signature = extractComponentSignature(
        resolver.checker,
        component
      );
      if (!signature) {
        return {
          props: UNKNOWN_TYPE,
          types: {}
        };
      }
      return analyze(logger, resolver, signature);
    }
  } : null;
}
function extractComponentSignature(checker, node) {
  const type = checker.getTypeAtLocation(node);
  for (const callSignature of type.getCallSignatures()) {
    if (isValidComponentReturnType(callSignature.getReturnType())) {
      return callSignature;
    }
  }
  if (type.symbol) {
    const classType = checker.getTypeOfSymbolAtLocation(type.symbol, node);
    for (const constructSignature of classType.getConstructSignatures()) {
      const returnType = constructSignature.getReturnType();
      if (returnType.getProperty("render")) {
        return constructSignature;
      }
    }
  }
  return null;
}
function isValidComponentReturnType(type) {
  if (isJsxElement(type)) {
    return true;
  }
  return false;
}
const jsxElementTypes = /* @__PURE__ */ new Set(["Element", "VNode"]);
function isJsxElement(type) {
  if (type.isUnion()) {
    for (const subtype of type.types) {
      if (isJsxElement(subtype)) {
        return true;
      }
    }
  }
  return jsxElementTypes.has(type.symbol?.getEscapedName().toString());
}

const PREACT_SPECIAL_TYPES = {
  Component: NODE_TYPE,
  ComponentClass: functionType(NODE_TYPE),
  ComponentChildren: NODE_TYPE,
  FunctionComponent: functionType(NODE_TYPE),
  FunctionalComponent: functionType(NODE_TYPE),
  PreactElement: NODE_TYPE,
  PreactNode: NODE_TYPE,
  PreactNodeArray: arrayType(NODE_TYPE),
  PreactPortal: NODE_TYPE
};

const preactFrameworkPlugin = {
  info: {
    apiVersion: 5,
    name: "@previewjs/plugin-preact"
  },
  isCompatible: async (dependencies) => {
    const version = await dependencies["preact"]?.readInstalledVersion();
    if (!version) {
      return false;
    }
    return parseInt(version) >= 10;
  },
  async create({ rootDir, reader, logger }) {
    const __dirname = url.fileURLToPath(new URL(".", import.meta.url));
    const previewDirPath = path.resolve(__dirname, "..", "preview");
    const typeAnalyzer = createTypeAnalyzer({
      rootDir,
      reader,
      specialTypes: PREACT_SPECIAL_TYPES,
      tsCompilerOptions: {
        jsx: ts.JsxEmit.ReactJSX,
        jsxImportSource: "preact",
        jsxFactory: "h",
        jsxFragmentFactory: "Fragment"
      }
    });
    return {
      defaultWrapperPath: "__previewjs__/Wrapper.tsx",
      previewDirPath,
      typeAnalyzer,
      crawlFiles: async (filePaths) => {
        const absoluteFilePaths = filePaths.map(
          (f) => path.isAbsolute(f) ? f : path.join(rootDir, f)
        );
        const resolver = typeAnalyzer.analyze(absoluteFilePaths);
        const components = [];
        const stories = [];
        for (const absoluteFilePath of absoluteFilePaths) {
          for (const previewable of await crawlFile(
            logger,
            resolver,
            rootDir,
            absoluteFilePath
          )) {
            if ("exported" in previewable) {
              components.push(previewable);
            } else {
              stories.push(previewable);
            }
          }
        }
        return {
          components,
          stories
        };
      },
      viteConfig: (configuredPlugins) => {
        return {
          resolve: {
            alias: {
              react: "preact/compat",
              "react-dom/test-utils": "preact/test-utils",
              "react-dom": "preact/compat",
              "react/jsx-runtime": "preact/jsx-runtime"
            }
          },
          esbuild: {
            jsx: "automatic",
            jsxImportSource: "preact",
            jsxFactory: "h",
            jsxFragment: "fragment"
          },
          plugins: [
            ...configuredPlugins.filter(
              (plugin) => plugin.name !== "prefresh" && plugin.name !== "vite:preact-jsx"
            ),
            {
              name: "previewjs:optimize-deps",
              config: () => ({
                optimizeDeps: {
                  include: ["preact", "preact/jsx-runtime"]
                }
              })
            }
          ]
        };
      },
      dispose: () => {
        typeAnalyzer.dispose();
      }
    };
  }
};

export { preactFrameworkPlugin as default };
